package org.processmining.niektax.attributebasedrefinement.dialog;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.niektax.attributebasedrefinement.plugin.ManualAttributeRefinementParameters;
import org.processmining.niektax.temporalrelations.model.Boundary;

public class ManualAttributeRefinementDialog extends JPanel implements ActionListener{

	private static final long serialVersionUID = -7564969692362738627L;
	ManualAttributeRefinementParameters configuration;
	
	JRadioButton conceptNameToggle;
	JRadioButton conceptNameLifeCycleToggle;

	JComboBox<String> activitySelector;
	JComboBox<String> attributeSelector;

	JTextField leftBoundTextField;
	JTextField rightBoundTextField;
	JButton addBoundsButton;
	
	JLabel areasLabel;
	JLabel boundaries;
	
	XLog log;
	String cnameConnector;
		
	public ManualAttributeRefinementDialog(XLog log, ManualAttributeRefinementParameters configuration, String cnameConnector){
		this.cnameConnector = cnameConnector;
		this.configuration = configuration;
		this.log = log;
		XConceptExtension   xce = XConceptExtension.instance();
		XLifecycleExtension xle = XLifecycleExtension.instance();
		Set<String> classifiedNames = new TreeSet<String>();
		
		JPanel container3 = new JPanel();
		JLabel classifierLabel = new JLabel("Select classifier: ");
		ButtonGroup group = new ButtonGroup();
		conceptNameLifeCycleToggle = new JRadioButton("Concept name + Lifecycle");
		conceptNameLifeCycleToggle.setSelected(true);
		conceptNameToggle = new JRadioButton("Concept name");
		group.add(conceptNameToggle);
		group.add(conceptNameLifeCycleToggle);
		conceptNameToggle.addActionListener(this);
		conceptNameLifeCycleToggle.addActionListener(this);
		container3.add(classifierLabel);
		container3.add(conceptNameToggle);
		container3.add(conceptNameLifeCycleToggle);
		this.add(container3);
		
		for(XTrace trace : log){
			for(XEvent xe : trace){
				String classifiedName = xce.extractName(xe) + cnameConnector + xle.extractTransition(xe);
				classifiedNames.add(classifiedName);
			}
		}
		
		activitySelector = new JComboBox<String>();
		
		configuration.setFirstActivity((String)classifiedNames.toArray()[0]);
		for(String classifiedName : classifiedNames)
			activitySelector.addItem(classifiedName);
		this.setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		activitySelector.addActionListener(this);
		JPanel container = new JPanel();
		container.add(activitySelector);
		JLabel attributeName = new JLabel("attribute name:");
		attributeSelector  = new JComboBox<String>();
		attributeSelector.addActionListener(this);
		container.add(attributeName);
		container.add(attributeSelector);
		this.add(container);
		
		JPanel container2 = new JPanel();
		JLabel leftBound = new JLabel("left bound:");
		leftBoundTextField  = new JTextField("",30);
		JLabel rightBound = new JLabel("right bound:");
		rightBoundTextField = new JTextField("",30);
		container2.add(leftBound);
		container2.add(leftBoundTextField);
		container2.add(rightBound);
		container2.add(rightBoundTextField);
		this.add(container2);
		JLabel boundExplanation = new JLabel("Left bound included, right bound excluded");
		this.add(boundExplanation);
		addBoundsButton = new JButton("Add bounds");
		addBoundsButton.addActionListener(this);
		this.add(addBoundsButton);
		boundaries = new JLabel("");
		this.add(boundaries);
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		if(e.getSource()==activitySelector){
			String activityName = (String) activitySelector.getSelectedItem();
			configuration.setFirstActivity(activityName);
			List<String> activityGlobalAttributes = calculateActivityGlobalAttributes(activityName);
			attributeSelector.removeAllItems();
			for(String attr : activityGlobalAttributes)
				attributeSelector.addItem(attr);
		}else if(e.getSource()==attributeSelector){
			String attributeName = (String) attributeSelector.getSelectedItem();
			configuration.setAttributeName(attributeName);
		}else if(e.getSource()==addBoundsButton){
			configuration.addBoundary(new Boundary(Double.valueOf(leftBoundTextField.getText()), Double.valueOf(rightBoundTextField.getText())));
			updateAndReset();
		}else if(e.getSource()==conceptNameToggle){
			activitySelector.removeAllItems();
			XConceptExtension   xce = XConceptExtension.instance();
			Set<String> classifiedNames = new TreeSet<String>();
			for(XTrace trace : log){
				for(XEvent xe : trace){
					String classifiedName = xce.extractName(xe);
					classifiedNames.add(classifiedName);
				}
			}
			for(String classifiedName : classifiedNames)
				activitySelector.addItem(classifiedName);
		}else if(e.getSource()==conceptNameLifeCycleToggle){
			activitySelector.removeAllItems();
			XConceptExtension   xce = XConceptExtension.instance();
			XLifecycleExtension xle = XLifecycleExtension.instance();
			Set<String> classifiedNames = new TreeSet<String>();
			for(XTrace trace : log){
				for(XEvent xe : trace){
					String classifiedName = xce.extractName(xe) + cnameConnector + xle.extractTransition(xe);
					classifiedNames.add(classifiedName);
				}
			}
			for(String classifiedName : classifiedNames)
				activitySelector.addItem(classifiedName);
		}
	}
	
	private List<String> calculateActivityGlobalAttributes(String attributeName){
		List<String> activityGlobalAttrs = new ArrayList<String>();
		boolean first = true;
		XConceptExtension   xce = XConceptExtension.instance();
		XLifecycleExtension xle = XLifecycleExtension.instance();
		for(XTrace trace : log){
			for(XEvent event : trace){
				String classifiedName = "";
				if(conceptNameLifeCycleToggle.isSelected())
					classifiedName = xce.extractName(event) + cnameConnector + xle.extractTransition(event);
				else if (conceptNameToggle.isSelected())
					classifiedName = xce.extractName(event);
					
				if(classifiedName.equals(attributeName)){
					XAttributeMap xam = event.getAttributes();
					if(first){
						activityGlobalAttrs.addAll(xam.keySet());
						first = false;
					}else{
						Set<String> removeSet = new HashSet<String>();
						for(String attr : activityGlobalAttrs){
							if(!xam.containsKey(attr))
								removeSet.add(attr);
						}
						activityGlobalAttrs.removeAll(removeSet);
					}
				}
			}
		}
		return activityGlobalAttrs;
	}
	
	private void updateAndReset(){
		leftBoundTextField.setText("");
		rightBoundTextField.setText("");
		boundaries.setText((configuration).getAreas().toString());
	}

	public String getCnameConnector(){
		return cnameConnector;
	}
}
