package org.processmining.niektax.attributebasedrefinement.plugin;


import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.commons.math3.util.Pair;
import org.processmining.niektax.temporalrelations.model.Boundary;
import org.processmining.niektax.temporalrelations.model.UnorderedPair;
import org.processmining.niektax.temporalrelations.model.dissimilarity.DissimilarityResult;

public class AttributeSplitBoundaries extends LinkedList<Boundary>{
	/*
	 * Left border of a split boundary is included, right border is not
	 */
	private static final long serialVersionUID = 8232466370306085632L;
	protected String activity;
	private TreeMap<Boundary, Double> dissimilarityScores;
	private HashMap<UnorderedPair<Boundary>, Set<DissimilarityResult>> splitReasons;
	
	public AttributeSplitBoundaries(String activity){
		this.activity = activity;
		this.dissimilarityScores = new TreeMap<Boundary,Double>();
		this.splitReasons = new HashMap<UnorderedPair<Boundary>, Set<DissimilarityResult>>();
	}
	
	public void addSplitReason(UnorderedPair<Boundary> boundaryPair, DissimilarityResult hypothesis){
		Set<DissimilarityResult> currentHypotheses = splitReasons.get(boundaryPair);
		if(currentHypotheses==null){
			currentHypotheses = new HashSet<DissimilarityResult>();
		}
	}
	
	public String getActivity(){
		return activity;
	}
	
	public void setDissimilarity(int i, double dissimilarity){
		dissimilarityScores.put(this.get(i), dissimilarity);
	}
	
	public Double getDisimilarity(int i){
		return dissimilarityScores.get(this.get(i));
	}
	
	public double getHighestDissimilarity(){
		double highest = 0.0;
		for (Entry<Boundary, Double> entry  : dissimilarityScores.entrySet()){
			if(entry.getValue()>highest)
				highest = entry.getValue();
		}
		return highest;
	}
	
	public boolean overlapping(int i, int j){
    	boolean nonOverlapping = true;
		// if bounds contained within existing bounds
    	if( (get(i).getLeft() >= get(j).getLeft() && get(i).getLeft() < get(j).getRight()) || (get(i).getRight() > get(j).getLeft() && get(i).getRight() <= get(j).getRight()) ){
    		nonOverlapping = false;
    	}
    	// if bounds contained within existing bounds
    	if( (get(i).getLeft() <= get(j).getLeft() && get(i).getRight() >= get(j).getRight()) || (get(i).getRight() <= get(j).getLeft() && get(i).getLeft() >= get(j).getRight()) ){
    		nonOverlapping = false;
    	}
    	return !nonOverlapping;
	}
	
	public Set<Boundary> getAsSet(){
		@SuppressWarnings("unchecked")
		List<Boundary> copyOfThis = (List<Boundary>) this.clone();
		Set<Boundary> resultSet = new TreeSet<Boundary>(copyOfThis);
		return resultSet;
	}
	
	public Set<Boundary> getMostDissimilarNonOverlappingBoundaries(double threshold){
		// TODO: check if additions match dissimilarity threshold
		Set<Boundary> resultSet = new TreeSet<Boundary>();
		for (Entry<Boundary, Double> entry  : entriesSortedByValues(dissimilarityScores)) {
		    boolean nonOverlapping = true;
			for(Boundary resultSetEntry : resultSet){
		    	Boundary curBound = entry.getKey();
		    	
		    	// if bounds contained within existing bounds
		    	if( (curBound.getLeft() >= resultSetEntry.getLeft() && curBound.getLeft() < resultSetEntry.getRight()) || (curBound.getRight() > resultSetEntry.getLeft() && curBound.getRight() <= resultSetEntry.getRight()) ){
		    		nonOverlapping = false;
		    	}
		    	// if bounds contained within existing bounds
		    	if( (curBound.getLeft() <= resultSetEntry.getLeft() && curBound.getRight() >= resultSetEntry.getRight()) || (curBound.getRight() <= resultSetEntry.getLeft() && curBound.getLeft() >= resultSetEntry.getRight()) ){
		    		nonOverlapping = false;
		    	}
		    }
			if(nonOverlapping && entry.getValue()>=threshold)
				resultSet.add(entry.getKey());
		}
		return resultSet;
	}
	
	public boolean dayTimeWithin(double value){
		boolean within = false;
		for(Boundary area : this){
			if(area.getLeft()<area.getRight()){
				if((value >= area.getLeft() && value < area.getRight())){
					within = true;
					break;
				}	
			}else{
				if(area.getLeft() > area.getRight()){
					if(value >= area.getLeft() ||value < area.getRight()){
						within = true;
						break;
					}
				}else{
					if (area.getLeft()==value){
						within = true;
						break;
					}
				}
			}
		}
		return within;
	}
	
	public boolean within(double value){
		boolean within = false;
		for(Boundary area : this){
			if(value >= area.getLeft() && value < area.getRight()){
				within = true;
				break;
			}
		}
		return within;
	}
	
	@Override
	public String toString(){
		String returnString = activity+": [";
		for(Boundary pair: this){
			returnString += "from "+pair.getLeft()+" to "+pair.getRight()+", ";
		}
		returnString += "]";
		return returnString;
	}
	
	public boolean sameSide(double value1, double value2){
		boolean within = true;
		for(Boundary area : this){
			if( (value1 <= area.getLeft() && value2 >= area.getRight()) || (value1 >= area.getLeft() && value2 <= area.getRight()) ){
				within = false;
				break;
			}
		}
		return within;
	}
	
	public Pair<Double,Double> fixOneSideOverlap(double left, double right){
		double correctionLeft  = 0.0;
		double correctionRight = 0.0;
		if( (within(left)&&!within(right)) || (within(right)&&!within(left))){
			for(Boundary area : this){
				if(left >= area.getLeft() && left <= area.getRight()){
					correctionLeft = Math.max(correctionLeft, area.getRight()-left);
				}
				if(right >= area.getLeft() && right <= area.getRight()){
					correctionRight = Math.max(correctionRight, area.getLeft()-right);
				}
			}
		}

		return new Pair<Double, Double>(correctionLeft, correctionRight);
	}
	
	static <K,V extends Comparable<? super V>> SortedSet<Map.Entry<K,V>> entriesSortedByValues(Map<K,V> map) {
        SortedSet<Map.Entry<K,V>> sortedEntries = new TreeSet<Map.Entry<K,V>>(
            new Comparator<Map.Entry<K,V>>() {
                @Override public int compare(Map.Entry<K,V> e1, Map.Entry<K,V> e2) {
                    int res = e1.getValue().compareTo(e2.getValue());
                    return res != 0 ? res : 1; // Special fix to preserve items with equal values
                }
            }
        );
        sortedEntries.addAll(map.entrySet());
        return sortedEntries;
    }
}
