package org.processmining.niektax.attributebasedrefinement.plugin;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.log.utils.XUtils;
import org.processmining.niektax.temporalrelations.model.Boundary;
import org.processmining.niektax.temporalrelations.model.CalculateLogEntropy;
import org.processmining.niektax.temporalrelations.model.TemporalRelationUtils;
import org.processmining.niektax.temporalrelations.model.UnorderedPair;
import org.processmining.niektax.temporalrelations.model.dissimilarity.DissimilarityMetrics;
import org.processmining.niektax.temporalrelations.model.dissimilarity.DissimilarityResult;

import com.google.common.collect.Table;

public class AttributeSplitLogBySplitBoundaries {
	private XLog originalLog;
	private XLog refinedLog;
	private String splitAttribute;
	private boolean IGCalculated;
	private double informationGain;
	private AttributeSplitBoundaries[] splitBoundariesArray;
	private Map<UnorderedPair<Boundary>, Set<DissimilarityResult>> dissimilarityResultMap;
	private Map<String, Set<String>> oldToNewActivityMap;
	private String cnameConnector;
	
	public AttributeSplitLogBySplitBoundaries(XLog log, AttributeSplitBoundaries[] splitBoundariesArray, boolean filterOutsideBounds, String splitAttribute, String cnameConnector){
		this.originalLog = log;
		this.refinedLog = XUtils.cloneLogWithoutClassifier(log);
		this.splitAttribute = splitAttribute;
		this.oldToNewActivityMap = new HashMap<String, Set<String>>();
		this.splitBoundariesArray = splitBoundariesArray;
		this.dissimilarityResultMap = new HashMap<UnorderedPair<Boundary>, Set<DissimilarityResult>>();
		this.IGCalculated = false;
		this.cnameConnector = cnameConnector;
		
		XConceptExtension   xce = XConceptExtension.instance();
		XLifecycleExtension xle = XLifecycleExtension.instance();
		HashSet<String> destinationActivities = new HashSet<String>();
	
		for(AttributeSplitBoundaries sb : splitBoundariesArray){
			destinationActivities.add(sb.getActivity());
		}
		
		for(XTrace t : refinedLog){
			for(XEvent xe : t){
				String activity  = xce.extractName(xe);
				String lifecycle = xle.extractTransition(xe);
				String classifiedName = activity+"+"+lifecycle;
				XAttributeMap xam = xe.getAttributes();
				if(!xam.containsKey(splitAttribute))
					continue;
				Double attributeValue = Double.parseDouble(xe.getAttributes().get(splitAttribute).toString());
				
				if(destinationActivities.contains(classifiedName)){
					AttributeSplitBoundaries splitter = null;
					for(AttributeSplitBoundaries sb : splitBoundariesArray){
						if(sb.getActivity().equals(classifiedName)){
							splitter = sb;
							break;
						}
					}
					Collections.sort(splitter);
					if(splitter.within(attributeValue)){
						for(int i=0; i<splitter.size(); i++){
							Boundary b = splitter.get(i);
							if(attributeValue>=b.getLeft() && attributeValue<b.getRight()){
								xce.assignName(xe, activity+"_"+(i+1));
								Set<String> current = oldToNewActivityMap.get(classifiedName);
								if(current==null)
									current = new TreeSet<String>();
								current.add(activity+"_"+(i+1)+"+"+lifecycle);
								oldToNewActivityMap.put(classifiedName, current);
							}
						}
					}else{
						xce.assignName(xe, activity+"_undefined");
						Set<String> current = oldToNewActivityMap.get(classifiedName);
						if(current==null)
							current = new HashSet<String>();
						current.add(activity+"_undefined"+"+"+lifecycle);
					}
				}
			}
		}
	}
	
	public double getEntropyLoss(){
		return CalculateLogEntropy.runHeadless(originalLog, false) - CalculateLogEntropy.runHeadless(refinedLog, false);
	}
	
	public double getLogRelationBasedRelativeInformationGain(){
		if(IGCalculated)
			return informationGain;
		
		Map<String, Table<String, String, Integer>> originalRelations = TemporalRelationUtils.calculateLogRelations(originalLog, cnameConnector);
		Map<String, Table<String, String, Integer>> newRelations = TemporalRelationUtils.calculateLogRelations(refinedLog, cnameConnector);
		
		Map<String, Integer> originalCounts = TemporalRelationUtils.getCounts(originalLog, cnameConnector);
		Map<String, Integer> newCounts = TemporalRelationUtils.getCounts(refinedLog, cnameConnector);
		
		double totalRelativeInformationGain = 0d;
		
		for(String key : oldToNewActivityMap.keySet()){
			Integer count = originalCounts.get(key);
			boolean partialSplit = newCounts.containsKey(key);
			List<String> newActivities = new LinkedList<String>(oldToNewActivityMap.get(key));
			Double oldEntropy = 0d;
			Map<String, Double> entropyByNewActivity = new HashMap<String, Double>();
			
			/*System.out.println("Totals");
			System.out.println(key+": "+count);
			for(String newEvent : newActivities)
				System.out.println(newEvent+": "+newCounts.get(newEvent));
			*/
			// ep
			//System.out.println("EP");
			if(originalRelations.containsKey("ep")){
				Map<String, Integer> originalEP = originalRelations.get("ep").row(key);
				//System.out.println(key+": "+originalEP);
				for(String other : originalEP.keySet()){
					Integer yes = originalEP.get(other);
					oldEntropy = updateEntropy(oldEntropy, yes, count);
				}
				for(String newEvent : newActivities){
					Map<String, Integer> newEP = newRelations.get("ep").row(newEvent);
					//System.out.println(newEvent+": "+newEP);
					Integer countN = newCounts.get(newEvent);
					for(String other : newEP.keySet()){
						Integer yes = newEP.get(other);
						Double currentEntropy = entropyByNewActivity.get(newEvent);
						if(currentEntropy==null)
							currentEntropy = 0d;
						currentEntropy = updateEntropy(currentEntropy, yes, countN);
						entropyByNewActivity.put(newEvent, currentEntropy);
					}
				}
				if(partialSplit){
					Map<String, Integer> newEP = newRelations.get("ep").row(key);
					//System.out.println(newEvent+": "+newEP);
					Integer countN = newCounts.get(key);
					for(String other : newEP.keySet()){
						Integer yes = newEP.get(other);
						Double currentEntropy = entropyByNewActivity.get(key);
						if(currentEntropy==null)
							currentEntropy = 0d;
						currentEntropy = updateEntropy(currentEntropy, yes, countN);
						entropyByNewActivity.put(key, currentEntropy);
					}
				}
			}
			// dp
			//System.out.println("DP");
			if(originalRelations.containsKey("dp")){
				Map<String, Integer> originalDP = originalRelations.get("dp").row(key);
				//System.out.println(key+": "+originalDP);
				for(String other : originalDP.keySet()){
					Integer yes = originalDP.get(other);
					oldEntropy = updateEntropy(oldEntropy, yes, count);
				}
				for(String newEvent : newActivities){
					Map<String, Integer> newDP = newRelations.get("dp").row(newEvent);
					//System.out.println(newEvent+": "+newDP);
					Integer countN = newCounts.get(newEvent);
					for(String other : newDP.keySet()){
						Integer yes = newDP.get(other);
						Double currentEntropy = entropyByNewActivity.get(newEvent);
						if(currentEntropy==null)
							currentEntropy = 0d;
						currentEntropy = updateEntropy(currentEntropy, yes, countN);
						entropyByNewActivity.put(newEvent, currentEntropy);
					}
				}
				if(partialSplit){
					Map<String, Integer> newDP = newRelations.get("dp").row(key);
					//System.out.println(newEvent+": "+newDP);
					Integer countN = newCounts.get(key);
					for(String other : newDP.keySet()){
						Integer yes = newDP.get(other);
						Double currentEntropy = entropyByNewActivity.get(key);
						if(currentEntropy==null)
							currentEntropy = 0d;
						currentEntropy = updateEntropy(currentEntropy, yes, countN);
						entropyByNewActivity.put(key, currentEntropy);
					}
				}
			}
			
			// ef
			//System.out.println("EF");
			if(originalRelations.containsKey("ef")){
				Map<String, Integer> originalEF = originalRelations.get("ef").row(key);
				//System.out.println(key+": "+originalEF);
				for(String other : originalEF.keySet()){
					Integer yes = originalEF.get(other);
					oldEntropy = updateEntropy(oldEntropy, yes, count);
				}
				for(String newEvent : newActivities){
					Map<String, Integer> newEF = newRelations.get("ef").row(newEvent);
					//System.out.println(newEvent+": "+originalEF);
					Integer countN = newCounts.get(newEvent);
					for(String other : newEF.keySet()){
						Integer yes = newEF.get(other);
						Double currentEntropy = entropyByNewActivity.get(newEvent);
						if(currentEntropy==null)
							currentEntropy = 0d;
						currentEntropy = updateEntropy(currentEntropy, yes, countN);
						entropyByNewActivity.put(newEvent, currentEntropy);
					}
				}
				if(partialSplit){
					Map<String, Integer> newEF = newRelations.get("ef").row(key);
					//System.out.println(newEvent+": "+originalEF);
					Integer countN = newCounts.get(key);
					for(String other : newEF.keySet()){
						Integer yes = newEF.get(other);
						Double currentEntropy = entropyByNewActivity.get(key);
						if(currentEntropy==null)
							currentEntropy = 0d;
						currentEntropy = updateEntropy(currentEntropy, yes, countN);
					}
				}
			}
			
			// df
			//System.out.println("DF");
			if(originalRelations.containsKey("df")){
				Map<String, Integer> originalDF = originalRelations.get("df").row(key);
				//System.out.println(key+": "+originalDF);
				for(String other : originalDF.keySet()){
					Integer yes = originalDF.get(other);
					oldEntropy = updateEntropy(oldEntropy, yes, count);
				}
				for(String newEvent : newActivities){
					Map<String, Integer> newDF = newRelations.get("df").row(newEvent);
					//System.out.println(newEvent+": "+newDF);
					Integer countN = newCounts.get(newEvent);
					for(String other : newDF.keySet()){
						Integer yes = newDF.get(other);
						Double currentEntropy = entropyByNewActivity.get(newEvent);
						if(currentEntropy==null)
							currentEntropy = 0d;
						currentEntropy = updateEntropy(currentEntropy, yes, countN);
						entropyByNewActivity.put(newEvent, currentEntropy);
					}
				}
				if(partialSplit){
					Map<String, Integer> newDF = newRelations.get("df").row(key);
					//System.out.println(newEvent+": "+newDF);
					Integer countN = newCounts.get(key);
					for(String other : newDF.keySet()){
						Integer yes = newDF.get(other);
						Double currentEntropy = entropyByNewActivity.get(key);
						if(currentEntropy==null)
							currentEntropy = 0d;
						currentEntropy = updateEntropy(currentEntropy, yes, countN);
						entropyByNewActivity.put(key, currentEntropy);
					}
				}
			}
			/*
			double totalNewEntropy = 0d;
			for(String newEvent : entropyByNewActivity.keySet()){
				totalNewEntropy += entropyByNewActivity.get(newEvent);
			}
			*/
			//int totalCount = 0;
			double newEntropy = 0d;
			for(String newAct : entropyByNewActivity.keySet()){
				//int newCount = newCounts.get(newAct);
				//double newEntropyTemp = entropyByNewActivity.get(newAct);
				//newEntropy = (newEntropy*totalCount + newEntropyTemp*newCount) / (totalCount + newCount);
				//totalCount += newCount;
				newEntropy+=entropyByNewActivity.get(newAct);
			}
			System.out.println(key+"="+oldEntropy+": "+entropyByNewActivity+"="+newEntropy+", "+((oldEntropy-newEntropy)/oldEntropy));
			totalRelativeInformationGain += (oldEntropy-newEntropy)/oldEntropy;
		}
		IGCalculated = true;
		informationGain = totalRelativeInformationGain / oldToNewActivityMap.keySet().size();
		return informationGain;
	}
	
	public void printFisherBasedTestStatistic(){
		for(String oldActivity : oldToNewActivityMap.keySet()){
			Set<String> set = oldToNewActivityMap.get(oldActivity);
			String[] newActivities = set.toArray(new String[set.size()]);
			for(int i=0; i<newActivities.length; i++){
				String new1 = newActivities[i];
				for(int j=i+1; j<newActivities.length; j++){
					String new2 = newActivities[j];
					System.out.println(oldActivity+"-> {"+new1+", "+new2+"}: ");
					Set<DissimilarityResult> results = TemporalRelationUtils.calculateActivityDissimilarity(refinedLog, new1, new2, oldActivity, DissimilarityMetrics.FISHER_TEST, 0.05, 0.0, cnameConnector);
					System.out.println("Statistical tests satisfying tresholds: ");
					for(DissimilarityResult result : results)
						System.out.println("p-value: "+result.getPValue()+" o/r: "+result.getOddsRatio()+ " relation: "+result.getHypothesisType()+" other: "+result.getOtherActivity());
				}
			}
		}
	}
	
	public AttributeSplitBoundaries[] getSplitBoundariesArray(){
		return splitBoundariesArray;
	}
	
	public String getSplitAttribute(){
		return splitAttribute;
	}
	
	public void addDissimilarityResults(UnorderedPair<Boundary> key, DissimilarityResult val){
		Set<DissimilarityResult> currentVals = dissimilarityResultMap.get(key);
		if(currentVals == null)
			currentVals = new HashSet<DissimilarityResult>();
		currentVals.add(val);
		dissimilarityResultMap.put(key, currentVals);
	}
	
	public Set<DissimilarityResult> getDissimilarityResults(UnorderedPair<Boundary> key){
		return dissimilarityResultMap.get(key);
	}
	
	private static double updateEntropy(double oldEntropy, int yes, int count){
		double yesPart = 0;
		double noPart = 0;
		Integer no = count-yes;
		if(yes!=0)
			yesPart = -yes*log2( ((double) yes) / count);
		if(no!=0)
			noPart = -no*log2( ((double) no) / count);
		//System.out.println("H("+yes+","+no+")= "+(yesPart+noPart));
		return oldEntropy + yesPart + noPart;
	}
	
	private static double log2(double a){
		return Math.log(a)/Math.log(2);
	}
	
	public XLog getSplittedLog(){
		return refinedLog;
	}
}
