package org.processmining.niektax.attributebasedrefinement.plugin;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.niektax.attributebasedrefinement.dialog.ManualAttributeRefinementDialog;
import org.processmining.niektax.temporalrelations.model.Boundary;

@Plugin(
		name = "Manual Daytime Attribute-based Refinement", 
		parameterLabels = {"Event log"}, 
	    returnLabels = {"Refined Log"}, 
	    returnTypes = { XLog.class })
public class ManualDayTimeBasedRefinementPlugin {
	public static final boolean verbose = true;
	public static final String cnameConnector = "_";
	
	private XLog splitByManualBoundaries(PluginContext pc, ManualAttributeRefinementParameters configuration, XLog log){
		Map<String, AttributeSplitBoundaries> splitBoundaryTable = configuration.obtainSplitBoundaries();

		// Split labels into manual boundaries
		List<AttributeSplitBoundaries> splitBoundariesList = new ArrayList<AttributeSplitBoundaries>();
		for(String activityToSplit : splitBoundaryTable.keySet()){
			AttributeSplitBoundaries sb = new AttributeSplitBoundaries(activityToSplit);
			Set<Boundary> mdnob = splitBoundaryTable.get(activityToSplit).getAsSet();
			sb.addAll(mdnob);
			splitBoundariesList.add(sb);
		}
		
		// Re-merge 
		System.out.println("splitBoundariesList: "+splitBoundariesList);
		System.out.println("attributeName:        "+configuration.getAttributeName());
		DayTimeAttributeSplitLogBySplitBoundaries aslbb = new DayTimeAttributeSplitLogBySplitBoundaries(log, splitBoundariesList.toArray(new AttributeSplitBoundaries[splitBoundariesList.size()]), true, configuration.getAttributeName(), cnameConnector, verbose);
		//System.out.println("Log-global Entropy loss: "+aslbb.getEntropyLoss());
		System.out.println("Log-relation information gain: "+aslbb.getLogRelationBasedRelativeInformationGain());
		aslbb.printFisherBasedTestStatistic();
		return aslbb.getSplittedLog();
	}
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl")
	@PluginVariant(variantLabel = "Manual Daytime Attribute-based Refinement", requiredParameterLabels = { 0 })
	public XLog splitByManualBoundaries(UIPluginContext context, XLog log) {
		// Get the default configuration.
		ManualAttributeRefinementParameters configuration = new ManualAttributeRefinementParameters();
	    // Get a dialog for this configuration.
	    ManualAttributeRefinementDialog dialog = new ManualAttributeRefinementDialog(log, configuration, cnameConnector);
	    // Show the dialog. User can now change the configuration.
	    InteractionResult result = context.showWizard("Configure Attribute-based Refinement", true, true, dialog);
	    // User has close the dialog.
	    if (result == InteractionResult.FINISHED) {
			// Do the heavy lifting.
	    	return splitByManualBoundaries(context, configuration, log);
	    }
	    
		return null;
	}
}
