package org.processmining.niektax.attributebasedrefinement.plugin;

import java.nio.file.Paths;
import java.util.Calendar;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.niektax.temporalrelations.model.Boundary;
import org.rosuda.REngine.REXP;
import org.rosuda.REngine.REXPMismatchException;
import org.rosuda.REngine.REngineException;
import org.rosuda.REngine.Rserve.RConnection;
import org.rosuda.REngine.Rserve.RserveException;

@Plugin(
		name = "Obtain Time-based Label Refinements", 
		parameterLabels = {"Input Log"}, 
	    returnLabels = {"Output Log"}, 
	    returnTypes = { XLog.class }
		)
public class ObtainTimeBasedRefinement {
	public static final boolean verbose = false;
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl")
	@PluginVariant(variantLabel = "Obtain Time-based Label Refinements", requiredParameterLabels = {0})
	public XLog obtainBorders(PluginContext context, XLog log) {
		Set<String> uniqueCNames = new HashSet<String>();
		Double bestIG = 0d;
		DayTimeAttributeSplitLogBySplitBoundaries bestSplit = null;
		
		XConceptExtension   xce = XConceptExtension.instance();
		XLifecycleExtension xle = XLifecycleExtension.instance();
		XTimeExtension		xte = XTimeExtension.instance();
		String cnameConnector = "_";
		
		try {
			RConnection c = new RConnection();
			// prevent plotting
			c.eval("pdf(file = NULL)");
			// fix working directory
			String path = Paths.get("").toAbsolutePath().toString().replace("\\", "/");
			c.eval("setwd('"+path+"')");
			// fix imports
			c.eval("source('"+path+"/R/imports.R')");
			// build data.frame from log
			c.eval("mydata <- data.frame(cname = character(), timestamp = integer(), stringsAsFactors = FALSE)");
			for (XTrace trace : log){
				for (XEvent event : trace){
					String cname = xce.extractName(event) + cnameConnector + xle.extractTransition(event);
					Date time = xte.extractTimestamp(event);
					Calendar fromMidnight = Calendar.getInstance();
					fromMidnight.setTime(time);
					fromMidnight.set(Calendar.HOUR_OF_DAY, 0);
					fromMidnight.set(Calendar.MINUTE, 0);
					fromMidnight.set(Calendar.SECOND, 0);
					fromMidnight.set(Calendar.MILLISECOND, 0);
					long timediff = (time.getTime() - fromMidnight.getTimeInMillis())/1000;
					c.eval("mydata[nrow(mydata)+1,] <- c('"+cname+"',"+timediff+")");
					uniqueCNames.add(cname);
				}
			}
			System.out.println("Start");
			for(String cname : uniqueCNames){
				System.out.println();
				if(cname.equals("9"+cnameConnector+"start")||cname.equals("23"+cnameConnector+"start"))
					continue;
				c.eval("antecedentEventSpecific"+cname+" <- mydata[mydata$cname == '"+cname+"',]");
				c.eval("antecedentEventStartSeconds"+cname+" <- antecedentEventSpecific"+cname+"$timestamp").asStrings();
				c.eval("source('"+path+"/R/rao.R')");
				c.eval("antecedentEventStartHours"+cname+" <- convertFunc(antecedentEventStartSeconds"+cname+")").asStrings();
				System.out.println(cname);
				c.eval("antecedentEventStartHours"+cname+" <- as.matrix(antecedentEventStartHours"+cname+")");
				c.eval("antecedentEventStartHoursAngle"+cname+" <- (360 * antecedentEventStartHours"+cname+") / 24");
				c.eval("antecedentEventStartHoursRadial"+cname+" <-  antecedentEventStartHoursAngle"+cname+" * pi / 180");

				REXP rao = c.eval("capture.output(rao.spacing.test(antecedentEventStartHoursRadial"+cname+", 0.01))"); // unfortunately we have to use capture.output here, as the critical value of is only printed to console and not part of the output object
				double rao_test_statistic = -1d;
				double rao_critical_value = -1d;
				for(String s : rao.asStrings()){
					if(!s.contains("="))
						continue;
					String[] ss = s.split("=");
					if(ss[0].contains("Test Statistic"))
						rao_test_statistic = Double.parseDouble(ss[1]);
					else if (ss[0].contains("critical value"))
						rao_critical_value = Double.parseDouble(ss[1]);
				}
				if(rao_critical_value>rao_test_statistic){
					System.out.println("No refinement for "+cname+", did not pass Rao's spacing test ("+rao_critical_value+">"+rao_test_statistic+")");
					continue;
				}
				
				REXP hartigan = c.eval("dip.test(antecedentEventStartHoursRadial"+cname+")");
				double hartigan_p_val = hartigan.asList().at(1).asDouble();
				if(hartigan_p_val>0.01){
					System.out.println("No refinement for "+cname+", did not pass Hartigan's dip test ("+hartigan_p_val+">0.01)");
					continue;
				}
				
				c.eval("x_val"+cname+" <- cos(antecedentEventStartHoursRadial"+cname+") * 1");
				c.eval("y_val"+cname+" <- sin(antecedentEventStartHoursRadial"+cname+") * 1");
				c.eval("final_matrix"+cname+" <- cbind(antecedentEventStartHours"+cname+", antecedentEventStartHoursAngle"+cname+", antecedentEventStartHoursRadial"+cname+", x_val"+cname+", y_val"+cname+")");
				c.eval("colnames(final_matrix"+cname+") <- c('hourFloat', 'hour_angle', 'hour_radii', 'x_coor', 'y_coor')");
				c.eval("wait"+cname+" <- as.matrix(final_matrix"+cname+"[, 4:5])");

				int maxComponents = 9;
				c.eval("set.seed(111)");
				
				String R_COMMAND_OR_SOURCE_FILE_PATH = "bic.mixvmf(wait"+cname+", A = "+maxComponents+")";
				REXP rResponseObject = null;

				try{
					rResponseObject = c.parseAndEval("a"+cname+" <- try(eval("+R_COMMAND_OR_SOURCE_FILE_PATH+"),silent=TRUE)"); //if(inherits(a, \"try-error\")){continue}else{a}");
					if(rResponseObject.inherits("try-error"))
						System.err.println("Unable to cluster activity "+cname+": "+rResponseObject.asString());
				}catch(REngineException e){
					System.err.println("Unable to cluster activity "+cname);
					continue;
				}
				
				double[] bics = rResponseObject.asList().at(0).asDoubles();
				int numComponents = 1;
				double minBic = Double.MAX_VALUE;
				for(int i=0; i<maxComponents; i++){
					if(bics[i]<(minBic-10)){
						minBic = bics[i];
						numComponents = i+1;
					}
				}
				System.out.println("Number of von Mises components for "+ cname +" according to BIC: "+numComponents);
				c.eval("wait"+cname+".vonMises <- movMF(wait"+cname+", "+numComponents+", control = list(E = 'softmax', nruns = 20,  kappa = list(common = TRUE, 'Banerjee_et_al_2005')))");
				c.eval("set.seed(2016)");
				boolean allSignificant = true;
				c.eval("wait"+cname+".vonMises.cluster <- predict(wait"+cname+".vonMises)");
				c.eval("matrixPosterior"+cname+" <- cbind(wait"+cname+".vonMises$P, final_matrix"+cname+", wait"+cname+".vonMises.cluster)");
				c.eval("mergedMatrix"+cname+" <- cbind(matrixPosterior"+cname+", antecedentEventSpecific"+cname+")");
				AttributeSplitBoundaries boundaries = new AttributeSplitBoundaries(cname);
				for(int i=1; i<=numComponents; i++){
					c.eval("colnames(mergedMatrix"+cname+")["+i+"] <- c('posterior"+i+"')");
					c.eval("cluster_mem"+cname+"_"+i+" <- subset(mergedMatrix"+cname+", wait"+cname+".vonMises.cluster == "+i+")");
					c.eval("cluster_mem"+cname+"_"+i+" <- cluster_mem"+cname+"_"+i+"[order(cluster_mem"+cname+"_"+i+"$posterior"+i+"),]");
					double[] hourFloats = c.eval("cluster_mem"+cname+"_"+i+"$hourFloat").asDoubles();
					double extreme1 = hourFloats[0];
					Double previous = null;
					Boolean previousDirectionUp = null;
					int directionChangeIndex = 0;
					for(int j=0; j<hourFloats.length; j++){
						double current = hourFloats[j];
						if(previous!=null && directionChangeIndex==0){
							if(previousDirectionUp==null){
								previousDirectionUp = current>previous;
							}else{
								if(previousDirectionUp){
									if(current<previous && current+24-previous>=9)
										directionChangeIndex = j-1;
								}else{
									if(current>previous && current+previous+24>=9)
										directionChangeIndex = j-1;
								}
							}
						}
						previous = current;
					}
					if(directionChangeIndex==0)
						directionChangeIndex=hourFloats.length-1;
					double extreme2 = hourFloats[directionChangeIndex];
					double minimum; double maximum;
					if(previousDirectionUp){
						minimum=extreme1; maximum=extreme2;
					}else{
						minimum=extreme2; maximum=extreme1;
					}
					//System.out.println("from: "+minimum+" to "+maximum);
					boundaries.add(new Boundary(minimum-0.1, maximum+0.1)); // add a tiny bit to upper bound, as upper bound is not included

					REXP watson = c.eval("capture.output(watson.test(cluster_mem"+cname+"_"+i+"$hour_radii, alpha = 0.01, dist = 'vonmises'))"); // unfortunately we have to use capture.output here, as the critical value of is only printed to console and not part of the output object
					double watson_test_statistic = -1d;
					double watson_critical_value = -1d;
					for(String s : watson.asStrings()){
						if(!s.contains(":"))
							continue;
						String[] ss = s.split(":");
						if(ss[0].contains("Test Statistic"))
							watson_test_statistic = Double.parseDouble(ss[1]);
						else if (ss[0].contains("critical value"))
							watson_critical_value = Double.parseDouble(ss[1]);
					}
					if(watson_critical_value>watson_test_statistic){
						// TODO: check goodness-of-fit test
						System.out.println("No refinement for "+cname+", component " + i + "did not pass Watson's goodness-of-fit test ("+watson_critical_value+">"+watson_test_statistic+")");
						allSignificant = false;
						continue;
					}
				}
				if(!allSignificant){
					System.out.println("No refinement for "+cname+", not all von Mises components fit the data well");
					continue;
				}else{
					System.out.println("clusters: "+boundaries.toString());
					DayTimeAttributeSplitLogBySplitBoundaries splitter = new DayTimeAttributeSplitLogBySplitBoundaries(log, new AttributeSplitBoundaries[]{boundaries}, false, "time:hourfloat", cnameConnector, verbose);
					double splitIG = splitter.getLogRelationBasedRelativeInformationGain();
					System.out.println("Information gain for splitting "+cname+": "+splitIG);
					if(splitIG>bestIG){
						bestIG = splitIG;
						bestSplit = splitter;
					}
				}
			}
		} catch (RserveException | REXPMismatchException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}

		return bestSplit.getSplittedLog();
	}
}
