package org.processmining.niektax.attributebasedrefinement.strategies;

import java.nio.file.Paths;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeContinuousImpl;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.niektax.attributebasedrefinement.plugin.AttributeSplitBoundaries;
import org.processmining.niektax.attributebasedrefinement.plugin.DayTimeAttributeSplitLogBySplitBoundaries;
import org.processmining.niektax.temporalrelations.model.Boundary;
import org.rosuda.REngine.REXP;
import org.rosuda.REngine.REXPMismatchException;
import org.rosuda.REngine.REngineException;
import org.rosuda.REngine.Rserve.RConnection;
import org.rosuda.REngine.Rserve.RserveException;

import com.google.common.collect.Sets;

@Plugin(
		name = "Exhaustive Information Gain Search for Best Refinements (by size)",
		parameterLabels = {"Input Log"},
	    returnLabels = {"Output Log"},
	    returnTypes = { XLog.class }
		)
public class ExhaustiveIGRefinementsOfSize {
	public static final boolean verbose = false;
	public static final int N = 3;

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl")
	@PluginVariant(variantLabel = "Exhaustive Information Gain Search for Best Refinements (by size)", requiredParameterLabels = {0})
	public XLog obtainBorders(PluginContext context, XLog log) {
		Set<String> uniqueCNames = new HashSet<String>();
		
		XConceptExtension   xce = XConceptExtension.instance();
		XLifecycleExtension xle = XLifecycleExtension.instance();
		XTimeExtension		xte = XTimeExtension.instance();
		String cnameConnector = "_";
		
		Map<AttributeSplitBoundaries, Double> scoredSplits = new HashMap<AttributeSplitBoundaries, Double>();
		RConnection c = null;
		try {
			c = new RConnection();
			// cleanup existing variables
			c.eval("rm(list=ls())");
			// prevent plotting
			c.eval("pdf(file = NULL)");
			// fix working directory
			String path = Paths.get("").toAbsolutePath().toString().replace("\\", "/");
			c.eval("setwd('"+path+"')");
			// fix imports
			c.eval("source('"+path+"/R/imports.R')");
			// build data.frame from log
			c.eval("mydata <- data.frame(cname = character(), timestamp = integer(), stringsAsFactors = FALSE)");
			for (XTrace trace : log){
				for (XEvent event : trace){
					String cname = xce.extractName(event) + cnameConnector + xle.extractTransition(event);
					Date time = xte.extractTimestamp(event);
					Calendar fromMidnight = Calendar.getInstance();
					fromMidnight.setTime(time);
					fromMidnight.set(Calendar.HOUR_OF_DAY, 0);
					fromMidnight.set(Calendar.MINUTE, 0);
					fromMidnight.set(Calendar.SECOND, 0);
					fromMidnight.set(Calendar.MILLISECOND, 0);
					long timediff = (time.getTime() - fromMidnight.getTimeInMillis())/1000;
					double hourfloat = (double) timediff / (60*60);
					event.getAttributes().put("time:hourfloat", new XAttributeContinuousImpl("time:hourfloat", hourfloat));
					c.eval("mydata[nrow(mydata)+1,] <- c('"+cname+"',"+timediff+")");
					uniqueCNames.add(cname);
				}
			}
			System.out.println("Start");
			for(String cname : uniqueCNames){
				c.eval("antecedentEventSpecific"+cname+" <- mydata[mydata$cname == '"+cname+"',]");
				c.eval("antecedentEventStartSeconds"+cname+" <- antecedentEventSpecific"+cname+"$timestamp").asStrings();
				c.eval("source('"+path+"/R/rao.R')");
				c.eval("antecedentEventStartHours"+cname+" <- convertFunc(antecedentEventStartSeconds"+cname+")").asStrings();
				System.out.println(cname);
				c.eval("antecedentEventStartHours"+cname+" <- as.matrix(antecedentEventStartHours"+cname+")");
				c.eval("antecedentEventStartHoursAngle"+cname+" <- (360 * antecedentEventStartHours"+cname+") / 24");
				c.eval("antecedentEventStartHoursRadial"+cname+" <-  antecedentEventStartHoursAngle"+cname+" * pi / 180");

				REXP rao = null;
				try {
					rao = c.parseAndEval("try(eval(capture.output(rao.spacing.test(antecedentEventStartHoursRadial"+cname+", 0.01))),silent=TRUE)");
					if (rao.inherits("try-error")) { System.out.println("rao: "+rao.asString()); }
				} catch (REngineException e1) {
					e1.printStackTrace();
				} // unfortunately we have to use capture.output here, as the critical value of is only printed to console and not part of the output object
				double rao_test_statistic = -1d;
				double rao_critical_value = -1d;
				for(String s : rao.asStrings()){
					if(!s.contains("="))
						continue;
					String[] ss = s.split("=");
					if(ss[0].contains("Test Statistic"))
						rao_test_statistic = Double.parseDouble(ss[1]);
					else if (ss[0].contains("critical value"))
						rao_critical_value = Double.parseDouble(ss[1]);
				}
				if(rao_critical_value>rao_test_statistic){
					System.out.println("No refinement for "+cname+", did not pass Rao's spacing test ("+rao_critical_value+">"+rao_test_statistic+")");
					continue;
				}
				
				REXP hartigan = c.eval("dip.test(antecedentEventStartHoursRadial"+cname+")");
				double hartigan_p_val = hartigan.asList().at(1).asDouble();
				if(hartigan_p_val>0.01){
					System.out.println("No refinement for "+cname+", did not pass Hartigan's dip test ("+hartigan_p_val+">0.01)");
					continue;
				}
				
				c.eval("x_val"+cname+" <- cos(antecedentEventStartHoursRadial"+cname+") * 1");
				c.eval("y_val"+cname+" <- sin(antecedentEventStartHoursRadial"+cname+") * 1");
				c.eval("final_matrix"+cname+" <- cbind(antecedentEventStartHours"+cname+", antecedentEventStartHoursAngle"+cname+", antecedentEventStartHoursRadial"+cname+", x_val"+cname+", y_val"+cname+")");
				c.eval("colnames(final_matrix"+cname+") <- c('hourFloat', 'hour_angle', 'hour_radii', 'x_coor', 'y_coor')");
				c.eval("wait"+cname+" <- as.matrix(final_matrix"+cname+"[, 4:5])");

				int maxComponents = 4;
				c.eval("set.seed(111)");
				
				String R_COMMAND_OR_SOURCE_FILE_PATH = "bic.mixvmf(wait"+cname+", A = "+maxComponents+")";
				REXP rResponseObject = null;

				try{
					rResponseObject = c.parseAndEval("a"+cname+" <- try(eval("+R_COMMAND_OR_SOURCE_FILE_PATH+"),silent=TRUE)"); //if(inherits(a, \"try-error\")){continue}else{a}");
					if(rResponseObject.inherits("try-error"))
						System.err.println("Unable to cluster activity "+cname+": "+rResponseObject.asString());
				}catch(REngineException e){
					System.err.println("Unable to cluster activity "+cname);
					continue;
				}
				
				double[] bics = rResponseObject.asList().at(0).asDoubles();
				int numComponents = 1;
				double minBic = Double.MAX_VALUE;
				for(int i=1; i<maxComponents; i++){
					if(bics[i]<(minBic-10)){
						minBic = bics[i];
						numComponents = i+1;
					}
				}
				System.out.println("Number of von Mises components for "+ cname +" according to BIC: "+numComponents);
				c.eval("wait"+cname+".vonMises <- movMF(wait"+cname+", "+numComponents+", control = list(E = 'softmax', nruns = 20,  kappa = list(common = TRUE, 'Banerjee_et_al_2005')))");
				c.eval("set.seed(2016)");
				boolean allSignificant = true;
				c.eval("wait"+cname+".vonMises.cluster <- predict(wait"+cname+".vonMises)");
				c.eval("matrixPosterior"+cname+" <- cbind(wait"+cname+".vonMises$P, final_matrix"+cname+", wait"+cname+".vonMises.cluster)");
				c.eval("mergedMatrix"+cname+" <- cbind(matrixPosterior"+cname+", antecedentEventSpecific"+cname+")");
				AttributeSplitBoundaries boundaries = new AttributeSplitBoundaries(cname);
				for(int i=1; i<=numComponents; i++){
					c.eval("colnames(mergedMatrix"+cname+")["+i+"] <- c('posterior"+i+"')");
					c.eval("cluster_mem"+cname+"_"+i+" <- subset(mergedMatrix"+cname+", wait"+cname+".vonMises.cluster == "+i+")");
					c.eval("cluster_mem"+cname+"_"+i+" <- cluster_mem"+cname+"_"+i+"[order(cluster_mem"+cname+"_"+i+"$posterior"+i+"),]");
					double[] hourFloats = c.eval("cluster_mem"+cname+"_"+i+"$hourFloat").asDoubles();
					double extreme1 = hourFloats[0];
					double maxDiff = 0d;
					double extreme2 = -1d;
					for(int j=1; j<hourFloats.length; j++){
						double current = hourFloats[j];
						double diff = Math.min(Math.abs(current-extreme1), Math.min(Math.abs(Math.abs(current-24-extreme1)), Math.abs(current+24-extreme1)));
						if(diff>maxDiff){
							maxDiff=diff;
							extreme2 = current;
						}
	
					}
					double minimum; double maximum;
					minimum = Math.min(extreme1, extreme2);
					maximum = Math.max(extreme1, extreme2);
					if((maximum-minimum) > (minimum+24-maximum)){
						double temp = minimum;
						minimum = maximum;
						maximum = temp;
					}
					boundaries.add(new Boundary(minimum-0.1, maximum+0.1)); // add a tiny bit to upper bound, as upper bound is not included

					REXP watson = c.eval("capture.output(watson.test(cluster_mem"+cname+"_"+i+"$hour_radii, alpha = 0.01, dist = 'vonmises'))"); // unfortunately we have to use capture.output here, as the critical value of is only printed to console and not part of the output object
					double watson_test_statistic = -1d;
					double watson_critical_value = -1d;
					for(String s : watson.asStrings()){
						if(!s.contains(":"))
							continue;
						String[] ss = s.split(":");
						if(ss[0].contains("Test Statistic"))
							watson_test_statistic = Double.parseDouble(ss[1]);
						else if (ss[0].contains("critical value"))
							watson_critical_value = Double.parseDouble(ss[1]);
					}
					if(watson_critical_value>watson_test_statistic){
						System.out.println("No refinement for "+cname+", component " + i + "did not pass Watson's goodness-of-fit test ("+watson_critical_value+">"+watson_test_statistic+")");
						allSignificant = false;
						continue;
					}
				}
				if(!allSignificant){
					System.out.println("No refinement for "+cname+", not all von Mises components fit the data well");
					continue;
				}else{
					System.out.println("clusters: "+boundaries.toString());
					DayTimeAttributeSplitLogBySplitBoundaries splitter = new DayTimeAttributeSplitLogBySplitBoundaries(log, new AttributeSplitBoundaries[]{boundaries}, false, "time:hourfloat", cnameConnector, verbose);
					if(splitter.lowestPValue()>0.01){
						System.out.println("No refinement for "+cname+", no significant difference found for clusters ("+splitter.lowestPValue()+">0.01)");
						continue;
					}
					double splitIG = splitter.getLogRelationBasedRelativeInformationGain();
					System.out.println("Split of "+cname+" p-value has: "+splitter.lowestPValue()+", and IG: "+splitIG);
					scoredSplits.put(boundaries, splitIG);
				}
			}
		} catch (RserveException | REXPMismatchException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		} finally{
			if(c!=null)
				c.close();
		}
		
		Set<Set<AttributeSplitBoundaries>> powerSet = Sets.powerSet(scoredSplits.keySet());

		XLog bestRefined = null;
		double bestIG = Double.NEGATIVE_INFINITY;
		System.out.println("powerset size: "+powerSet.size());
		for(Set<AttributeSplitBoundaries> refinementSet : powerSet){
			if(refinementSet.size()!=N)
				continue;
			DayTimeAttributeSplitLogBySplitBoundaries splitter = new DayTimeAttributeSplitLogBySplitBoundaries(log, refinementSet.toArray(new AttributeSplitBoundaries[refinementSet.size()]), false, "time:hourfloat", cnameConnector, verbose);
			// TODO: calculate fitness/precision/fscore of the IMf 20% model of this refined log
			XLog refLog = splitter.getSplittedLog();
			Double ig = splitter.getLogRelationBasedRelativeInformationGain();
			if(ig>bestIG){
				bestIG = ig;
				bestRefined = refLog;
			}
		}
		System.out.println();
		System.out.println("Best refinement | information gain: "+bestIG);
		return bestRefined;
	}
}
