package org.processmining.niektax.temporalrelations.model;

import org.processmining.niektax.temporalrelations.model.exceptions.SplitBoundariesNotBuiltException;


public class BoundaryCalculator {
	private SplitBoundaries splitBoundaries;
	private double tempLeftBoundary;
	private double tempLatest;
	private double dx;
	private boolean built;
	
	public BoundaryCalculator(double dx, String firstActivity, String secondActivity){
		splitBoundaries = new SplitBoundaries(firstActivity, secondActivity);
		this.tempLeftBoundary = -1;
		this.tempLatest = -1;
		this.dx = dx;
		this.built = false;
	}
	
	public void add(double value){
		if(tempLeftBoundary == -1){
			tempLeftBoundary = value;
			tempLatest = value;
		}else{
			if( Math.abs((tempLatest+dx)-value) < 1){
				tempLatest = value;
			}else{
				splitBoundaries.add(new Boundary(tempLeftBoundary, tempLatest));
				tempLeftBoundary = -1;
			}
		}
	}
	
	public void buildSplitBoundaries(){
		splitBoundaries.add(new Boundary(tempLeftBoundary, Double.MAX_VALUE));
		tempLeftBoundary = -1;
		built = true;
	}
	
	public SplitBoundaries getSplitBoundaries() throws SplitBoundariesNotBuiltException{
		if(built)
			return splitBoundaries;
		else
			throw new SplitBoundariesNotBuiltException();
	}
}
