package org.processmining.niektax.temporalrelations.model;

import java.util.HashMap;
import java.util.Map;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;

@Plugin(
		name = "Calculate Log Entropy (Std. Out)", 
		parameterLabels = {"Log"}, 
	    returnLabels = {"Same log"}, 
	    returnTypes = { XLog.class })
public class CalculateLogEntropy {
	private static String traceEndSymbol = "^";
	
	
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "N. Tax", email = "n.tax@tue.nl")
	@PluginVariant(variantLabel = "Calculate Log Entropy (Std. Out)", requiredParameterLabels = {0})
	public XLog calculateEntropy(PluginContext context, XLog log){
		runHeadless(log, true);
		return log;
	}
	
	public static Double runHeadless(XLog log, boolean verbose){
		Map<String, Integer> map = new HashMap<String, Integer>();
		// count the occurrences of each value
		int numEvents = 0;
		for(XTrace trace : log){
			for(XEvent event : trace){
				String sequence = event.getAttributes().get("concept:name").toString();
				numEvents++;
				if (!map.containsKey(sequence)) {
					map.put(sequence, 0);
				}
				map.put(sequence, map.get(sequence) + 1);
			}
			if (!map.containsKey(traceEndSymbol)) {
				map.put(traceEndSymbol, 0);
			}
			map.put(traceEndSymbol, map.get(traceEndSymbol) + 1);
		}

		// calculate the entropy
		Double result = 0.0;
		for (String eventType : map.keySet()) {
			Double frequency = (double) map.get(eventType) / numEvents;
			result -= frequency * (Math.log(frequency) / Math.log(2));
		}
		if(verbose)
			System.out.println("Log entropy: "+result);
		return result;
	}
	
}