package org.processmining.niektax.temporalrelations.model;

import java.util.Set;

import javax.swing.BoxLayout;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.SwingConstants;

import org.apache.commons.math3.stat.descriptive.DescriptiveStatistics;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.statistics.SimpleHistogramBin;
import org.jfree.data.statistics.SimpleHistogramDataset;

import com.google.common.collect.Multiset;
import com.google.common.collect.Table;

public class GenerateHistogramTileThread implements Runnable{
	Table<String, String, Multiset<Long>> timeRelations;
	Table<String, String, JPanel> panelTable;
	String row;
	String column;
	int rowIndex;
	int columnIndex;
	
	public GenerateHistogramTileThread(Table<String, String, Multiset<Long>> timeRelations, Table<String, String, JPanel> panelTable, String row, String column, int rowIndex, int columnIndex){
		this.timeRelations = timeRelations;
		this.panelTable    = panelTable;
		this.row           = row;
		this.column        = column;
		this.rowIndex      = rowIndex;
		this.columnIndex   = columnIndex;
	}
	
	private int maxValue(Set<Long> set){
		long maxValue = Long.MIN_VALUE;
		for(Long value : set)
			if(value>maxValue)
				maxValue=value;
		return (int) maxValue;
	}

	public void run() {
		JPanel returnPanel = new JPanel();
		returnPanel.setLayout(new BoxLayout(returnPanel, BoxLayout.Y_AXIS));

		// Headers
		JPanel innerPanel = new JPanel();
		innerPanel.setLayout(new BoxLayout(innerPanel, BoxLayout.Y_AXIS));
		if(columnIndex == 0){
			returnPanel.setLayout(new BoxLayout(returnPanel, BoxLayout.X_AXIS));
			JLabel label = new JLabel(row);
			label.setVerticalAlignment(SwingConstants.CENTER);
			returnPanel.add(label);
			returnPanel.add(innerPanel);
		}
		
		// Headers
		if(rowIndex == 0){
			JLabel label = new JLabel(column);
			label.setVerticalAlignment(SwingConstants.CENTER);
			if(columnIndex==0){
				innerPanel.add(label);
			}else{
				returnPanel.add(label);
			}
		}
		
		Multiset<Long> multisetData = timeRelations.get(row, column);
		
		if(multisetData != null){
			// Histogram
			int maxBins = 100;
			int bins = Math.min(maxValue(multisetData.elementSet()), maxBins);

			if(bins>0){
				int aggregationfactor = (int)Math.ceil((double)maxValue(multisetData.elementSet()) / bins);
				
				SimpleHistogramDataset dataset = new SimpleHistogramDataset(""+multisetData.size());
				for(int i=0; i<=bins; i++){
					dataset.addBin(new SimpleHistogramBin(i*aggregationfactor-0.5*aggregationfactor, i*aggregationfactor+0.5*aggregationfactor, false, true));
				}
				dataset.setAdjustForBinSize(false);
			
				DescriptiveStatistics ds = new DescriptiveStatistics();
				for(Long timediff : multisetData.elementSet()){
					for(int i=0; i<multisetData.count(timediff); i++){
						ds.addValue(timediff.doubleValue());
						try{
							dataset.addObservation(timediff.doubleValue());
						}catch(RuntimeException e){
							System.out.println("diff: "+timediff.doubleValue());
							throw e;
						}
					}
				}

				JFreeChart chart = ChartFactory.createHistogram(row+" - "+column, "Time difference", "Frequency", 
						dataset, PlotOrientation.VERTICAL, false, true, false);
				ChartPanel histogramPanel = new ChartPanel(chart);
				JLabel meanLabel		 	  = new JLabel("Mean:             "+ds.getMean());
				JLabel standardDeviationLabel = new JLabel("Std. Dev:         "+ds.getStandardDeviation());
				JLabel kurtosisLabel          = new JLabel("Kurtosis:         "+ds.getKurtosis());
				
				if(columnIndex==0){
					innerPanel.add(histogramPanel);
					innerPanel.add(meanLabel);
					innerPanel.add(standardDeviationLabel);
					innerPanel.add(kurtosisLabel);
				}else{
					returnPanel.add(histogramPanel);
					returnPanel.add(meanLabel);
					returnPanel.add(standardDeviationLabel);
					returnPanel.add(kurtosisLabel);
				}
			}
		}
		panelTable.put(row, column, returnPanel);
	}
}
