package org.processmining.niektax.temporalrelations.model;

import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import org.apache.commons.math3.util.Pair;
import org.processmining.niektax.temporalrelations.model.dissimilarity.DissimilarityResult;

public class SplitBoundaries extends LinkedList<Boundary>{
	/*
	 * Left area of a split boundary is included, right area is not
	 */
	private static final long serialVersionUID = 8232466370306085632L;
	protected String firstActivity;
	protected String secondActivity;
	private TreeMap<Boundary, Double> dissimilarityScores;
	private Map<UnorderedPair<Boundary>, Set<DissimilarityResult>> splitReasons;
	//private Set<DissimilarityResult> combinedDissimilarityResult;
	private Map<Boundary, Set<DissimilarityResult>> dissimilarity1OriginalResult;
	private Map<Boundary, Set<DissimilarityResult>> dissimilarity2OriginalResult;
	
	public SplitBoundaries(String firstActivity, String secondActivity){
		this.firstActivity  = firstActivity;
		this.secondActivity = secondActivity;
		this.dissimilarityScores = new TreeMap<Boundary,Double>();
		this.splitReasons = new HashMap<UnorderedPair<Boundary>, Set<DissimilarityResult>>();
		//this.combinedDissimilarityResult = new HashSet<DissimilarityResult>();
		this.dissimilarity1OriginalResult = new HashMap<Boundary, Set<DissimilarityResult>>();
		this.dissimilarity2OriginalResult = new HashMap<Boundary, Set<DissimilarityResult>>();
	}
	
	/*
	public void setCombinedDissimilarity(Set<DissimilarityResult> dr){
		combinedDissimilarityResult = dr;
	}
	
	public Set<DissimilarityResult> getCombinedDissimilarity(){
		return combinedDissimilarityResult;
	}
	
	public void removeFromCombinedDissimilarity(DissimilarityResult dr){
		combinedDissimilarityResult.remove(dr);
	}*/
	
	public void setDissimilarity1Original(Boundary b, Set<DissimilarityResult> dr){
		dissimilarity1OriginalResult.put(b, dr);
	}
	
	public Set<DissimilarityResult> getDissimilarity1Original(Boundary b){
		Set<DissimilarityResult> dr = dissimilarity1OriginalResult.get(b);
		if(dr==null)
			dr = new HashSet<DissimilarityResult>();
		return dr;
	}
	
	public void removeFromDissimilarity1Original(DissimilarityResult dr){
		dissimilarity1OriginalResult.remove(dr);
	}
	
	public void setDissimilarity2Original(Boundary b, Set<DissimilarityResult> dr){
		dissimilarity2OriginalResult.put(b, dr);
	}
	
	public Set<DissimilarityResult> getDissimilarity2Original(Boundary b){
		Set<DissimilarityResult> dr = dissimilarity2OriginalResult.get(b);
		if(dr==null)
			dr = new HashSet<DissimilarityResult>();
		return dr;
	}
	
	public void removeFromDissimilarity2Original(DissimilarityResult dr){
		dissimilarity2OriginalResult.remove(dr);
	}
	
	public void setSplitReason(UnorderedPair<Boundary> boundaryPair, Set<DissimilarityResult> hypothesis){
		splitReasons.put(boundaryPair, hypothesis);
	}
	
	public void addSplitReason(UnorderedPair<Boundary> boundaryPair, DissimilarityResult hypothesis){
		Set<DissimilarityResult> currentHypotheses = splitReasons.get(boundaryPair);
		if(currentHypotheses==null){
			currentHypotheses = new HashSet<DissimilarityResult>();
		}
		currentHypotheses.add(hypothesis);
		splitReasons.put(boundaryPair, currentHypotheses);
	}
	
	public Set<DissimilarityResult> getSplitReasons(UnorderedPair<Boundary> boundaryPair){
		Set<DissimilarityResult> current = splitReasons.get(boundaryPair);
		if(current==null)
			current = new HashSet<DissimilarityResult>();
		return current;
	}
	
	public void removeSplitReason(UnorderedPair<Boundary> boundaryPair, DissimilarityResult hypothesis){
		Set<DissimilarityResult> currentHypotheses = splitReasons.get(boundaryPair);
		if(currentHypotheses==null){
			currentHypotheses = new HashSet<DissimilarityResult>();
		}
		currentHypotheses.remove(hypothesis);
		splitReasons.put(boundaryPair, currentHypotheses);
	}
	
	public Map<UnorderedPair<Boundary>, Set<DissimilarityResult>> getAllSplitReasons(){
		return splitReasons;
	}
	
	public String getFirstActivity(){
		return firstActivity;
	}
	
	public String getSecondActivity(){
		return secondActivity;
	}
	
	public void setDissimilarity(int i, double dissimilarity){
		dissimilarityScores.put(this.get(i), dissimilarity);
	}
	
	public Double getDisimilarity(int i){
		return dissimilarityScores.get(this.get(i));
	}
	
	public double getHighestDissimilarity(){
		double highest = 0.0;
		for (Entry<Boundary, Double> entry  : dissimilarityScores.entrySet()){
			if(entry.getValue()>highest)
				highest = entry.getValue();
		}
		return highest;
	}
	
	public boolean overlapping(int i, int j){
    	boolean nonOverlapping = true;
		// if bounds contained within existing bounds
    	if( (get(i).getLeft() >= get(j).getLeft() && get(i).getLeft() < get(j).getRight()) || (get(i).getRight() > get(j).getLeft() && get(i).getRight() <= get(j).getRight()) ){
    		nonOverlapping = false;
    	}
    	// if bounds contained within existing bounds
    	if( (get(i).getLeft() <= get(j).getLeft() && get(i).getRight() >= get(j).getRight()) || (get(i).getRight() <= get(j).getLeft() && get(i).getLeft() >= get(j).getRight()) ){
    		nonOverlapping = false;
    	}
    	return !nonOverlapping;
	}
	
	public Set<Boundary> getAsSet(){
		@SuppressWarnings("unchecked")
		List<Boundary> copyOfThis = (List<Boundary>) this.clone();
		Set<Boundary> resultSet = new TreeSet<Boundary>(copyOfThis);
		return resultSet;
	}
	
	public Set<Boundary> getMostDissimilarNonOverlappingBoundaries(double threshold){
		// TODO: check if additions match dissimilarity threshold
		Set<Boundary> resultSet = new TreeSet<Boundary>();
		for (Entry<Boundary, Double> entry  : entriesSortedByValues(dissimilarityScores)) {
		    boolean nonOverlapping = true;
			for(Boundary resultSetEntry : resultSet){
		    	Boundary curBound = entry.getKey();
		    	
		    	// if bounds contained within existing bounds
		    	if( (curBound.getLeft() >= resultSetEntry.getLeft() && curBound.getLeft() < resultSetEntry.getRight()) || (curBound.getRight() > resultSetEntry.getLeft() && curBound.getRight() <= resultSetEntry.getRight()) ){
		    		nonOverlapping = false;
		    	}
		    	// if bounds contained within existing bounds
		    	if( (curBound.getLeft() <= resultSetEntry.getLeft() && curBound.getRight() >= resultSetEntry.getRight()) || (curBound.getRight() <= resultSetEntry.getLeft() && curBound.getLeft() >= resultSetEntry.getRight()) ){
		    		nonOverlapping = false;
		    	}
		    }
			if(nonOverlapping && entry.getValue()>=threshold)
				resultSet.add(entry.getKey());
		}
		return resultSet;
	}
	
	public boolean within(double value){
		boolean within = false;
		for(Boundary area : this){
			if(value >= area.getLeft() && value < area.getRight()){
				within = true;
				break;
			}
		}
		return within;
	}
	
	@Override
	public String toString(){
		String returnString = firstActivity + " to " + secondActivity+ ": [";
		for(Boundary pair: this){
			returnString += "from "+pair.getLeft()+" to "+pair.getRight()+", ";
		}
		returnString += "]";
		return returnString;
	}
	
	public boolean sameSide(double value1, double value2){
		boolean within = true;
		for(Boundary area : this){
			if( (value1 <= area.getLeft() && value2 >= area.getRight()) || (value1 >= area.getLeft() && value2 <= area.getRight()) ){
				within = false;
				break;
			}
		}
		return within;
	}
	
	public Pair<Double,Double> fixOneSideOverlap(double left, double right){
		double correctionLeft  = 0.0;
		double correctionRight = 0.0;
		if( (within(left)&&!within(right)) || (within(right)&&!within(left))){
			for(Boundary area : this){
				if(left >= area.getLeft() && left <= area.getRight()){
					correctionLeft = Math.max(correctionLeft, area.getRight()-left);
				}
				if(right >= area.getLeft() && right <= area.getRight()){
					correctionRight = Math.max(correctionRight, area.getLeft()-right);
				}
			}
		}

		return new Pair<Double, Double>(correctionLeft, correctionRight);
	}
	
	static <K,V extends Comparable<? super V>> SortedSet<Map.Entry<K,V>> entriesSortedByValues(Map<K,V> map) {
        SortedSet<Map.Entry<K,V>> sortedEntries = new TreeSet<Map.Entry<K,V>>(
            new Comparator<Map.Entry<K,V>>() {
                @Override public int compare(Map.Entry<K,V> e1, Map.Entry<K,V> e2) {
                    int res = e1.getValue().compareTo(e2.getValue());
                    return res != 0 ? res : 1; // Special fix to preserve items with equal values
                }
            }
        );
        sortedEntries.addAll(map.entrySet());
        return sortedEntries;
    }

}
