package org.processmining.niektax.temporalrelations.model;

import org.apache.commons.math3.analysis.UnivariateFunction;

import com.google.common.collect.Multiset;

import dr.math.distributions.NormalKDEDistribution;

public class TimeRelation implements UnivariateFunction{
	NormalKDEDistribution kde;
	double 			      silverman;
	double			      variance;
	double			      mean;
	double			      stddev;
	int			 	      n;
	boolean			      initialised;
	
	public TimeRelation(Multiset<Long> kdeInput){
		initialised = false;
		if(kdeInput != null){
			initialised = true;
			
			int  sample_size   = 20000;
			int reservoirSize  = Math.min(sample_size, kdeInput.size());
			Double[] diffReservoir = new Double[reservoirSize];
			if(kdeInput.size()>sample_size){
				// perform reservoir sampling
				int i = 0;
				while(i<diffReservoir.length){
					Long l = (long)(kdeInput.elementSet().toArray())[0];
					diffReservoir[i] = Double.valueOf(l);
					kdeInput.setCount(l, kdeInput.count(l) - 1);
					i++;
				}
				for(Long key : kdeInput.elementSet()){
					int count = kdeInput.count(key);
					int j =0;
					while(j<count){
						int r = (int) Math.floor(Math.random()*(i+j));
						if(r<reservoirSize){
							diffReservoir[r] = Double.valueOf(key);
						}
						j++;
					}
				}
			}else{
				int i = 0;
				for(Long key : kdeInput.elementSet()){
					for(int j=0; j<kdeInput.count(key); j++){
						diffReservoir[i] = Double.valueOf(key);
						i++;
					}
				}
			}
			
			// Calculate Silverman's rule of Thumb
			double sum = 0;
		    for (int i = 0; i < diffReservoir.length; i++){
		        sum += diffReservoir[i];
		    }
		    n = diffReservoir.length;
		    mean = sum/n;
		    double temp = 0;
		    for (int i = 0; i < n; i++)
	            temp += ((mean-diffReservoir[i])*(mean-diffReservoir[i]));
	        variance  = temp/n;
	        stddev    = Math.sqrt(variance);
		    silverman = 1.06*stddev*Math.pow(n, -(1/5));
		    //silverman = silverman / 100;
			System.out.println("Size: "+diffReservoir.length+"  Silverman: "+silverman+"  sum: "+sum+"  mean: "+mean+"  variance: "+variance+"  stddev: "+stddev);
			long before = System.currentTimeMillis();
			kde = new NormalKDEDistribution(diffReservoir, null, null, silverman);
			System.out.println("Kernel Density Estimator completed in "+((before-System.currentTimeMillis())/1000)+" seconds");
		}
	}
	
	public double value(double x){
		return kde.pdf(x);
	}
	
	public double getSilverman(){
		return silverman;
	}
	
	public double getMean(){
		return mean;
	}
	
	public double getVariance(){
		return variance;
	}
	
	public double getStddev(){
		return stddev;
	}
	
	public int getN(){
		return n;
	}
	
	public boolean isInitialised(){
		return initialised;
	}
}
