package org.processmining.niektax.temporalrelations.model;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.niektax.temporalrelations.model.TimeRelationsCalculator.LookBack;

import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;
import com.google.common.collect.Table;

public class TimeRelationsCalculatorThread implements Runnable{
	XTrace t;
	TimeRelationsCalculator.LookBack lookBack;
	Table<String, String, Multiset<Long>> timeRelations;
	Table<String, String, Long> maxValues;
	Table<String, String, Long> minValues;


	public TimeRelationsCalculatorThread(XTrace t, LookBack lookBack, Table<String, String, Multiset<Long>> timeRelations, Table<String, String, Long> maxValues, Table<String, String, Long> minValues){
		this.t = t;
		this.lookBack = lookBack;
		this.timeRelations = timeRelations;
	}
	
	public void run(){
		XTimeExtension      te = XTimeExtension.instance();
		XConceptExtension   ce = XConceptExtension.instance();
		XLifecycleExtension xl = XLifecycleExtension.instance();

		Map<String, Date> lastSeen = new HashMap<String, Date>();
		boolean firstEvent = true;
		
		for(XEvent e : t){
			String activity = ce.extractName(e);
			activity+="+"+xl.extractTransition(e);
			
			Date date = te.extractTimestamp(e);

			if(!firstEvent){
				Date lastSelf = lastSeen.get(activity);
				for(String str: lastSeen.keySet()){
					if(lastSeen.get(str)!=null){
						if(lookBack==LookBack.UNTIL_FIRST_SELF && lastSelf!=null && lastSeen.get(str).before(lastSelf))
							continue;
						long currentTimeDiff = ( date.getTime() - lastSeen.get(str).getTime() ) / (60*1000);
						synchronized(timeRelations){
							Multiset<Long> timeDifferences = timeRelations.get(str, activity);
							if(timeDifferences == null)
								timeDifferences = HashMultiset.create();
							timeDifferences.add(currentTimeDiff);
							timeRelations.put(str, activity, timeDifferences);
						}
					}
				}
			}
			lastSeen.put(activity, date);
			firstEvent = false;
		}
	}
}
