package org.processmining.niektax.temporalrelations.model.dissimilarity;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.math3.exception.NotStrictlyPositiveException;
import org.apache.commons.math3.stat.inference.ChiSquareTest;
import org.processmining.niektax.temporalrelations.model.dissimilarity.DissimilarityResult.HypothesisType;

import com.google.common.collect.Table;

public class ChiSquareTestDissimilarity extends DissimilarityMetric{
	
	public ChiSquareTestDissimilarity(String a, String b, String originalAb, Table<String, String, Integer> df, Table<String, String, Integer> ef, Table<String, String, Integer> ep, Table<String, String, Integer> cooc, Set<String> classifiedNames, Map<String, Integer> counts, double thresholdP, double thresholdOdds){
		super(a,b,originalAb,df,ef,ep,cooc,classifiedNames,counts,thresholdP,thresholdOdds);
	}
	
	public Set<DissimilarityResult> getParallelDissimilarity() {		
		// TODO Auto-generated method stub
		return null;
	}

	public Set<DissimilarityResult> getDirectlyFollowsDissimilarity() {
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();
		
		Map<String, Integer> aRow = df.row(a);
		Map<String, Integer> bRow = df.row(b);
		List<Long> observeda = new ArrayList<Long>();
		List<Long> observedb = new ArrayList<Long>();
		long countoa = 0;
		long countob = 0;
		for(String c : classifiedNames){
			long oa = aRow.get(c) == null ? 0 : aRow.get(c);
			long ob = bRow.get(c) == null ? 0 : bRow.get(c);
			if(oa>0 || ob>0){
				observeda.add(oa);
				observedb.add(ob);
				countoa += oa;
				countob += ob;
			}
		}
		ChiSquareTest cst = new ChiSquareTest();
		double returnValue = 0.0;
		if(countoa>0 && countob>0){
			try{
				returnValue = 1-cst.chiSquareTestDataSetsComparison(ArrayUtils.toPrimitive(observeda.toArray(new Long[observeda.size()])), ArrayUtils.toPrimitive(observedb.toArray(new Long[observedb.size()])));
			}catch(NotStrictlyPositiveException e){
				returnValue = 0.0;
			}
		}
		
		if(returnValue>thresholdP)
			results.add(new DissimilarityResult("UNKOWN", HypothesisType.DIRECT_FOLLOWS, returnValue, 0.0));
		
		return results;
	}

	public Set<DissimilarityResult> getDirectlyPrecedesDissimilarity() {
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();

		Map<String, Integer> aCol = df.column(a);
		Map<String, Integer> bCol = df.column(b);
		List<Long> observeda = new ArrayList<Long>();
		List<Long> observedb = new ArrayList<Long>();
		long countoa = 0;
		long countob = 0;
		for(String c : classifiedNames){
			long oa = aCol.get(c) == null ? 0 : aCol.get(c);
			long ob = bCol.get(c) == null ? 0 : bCol.get(c);
			if(oa>0 || ob>0){
				observeda.add(oa);
				observedb.add(ob);
				countoa += oa;
				countob += ob;
			}
		}
		ChiSquareTest cst = new ChiSquareTest();
		double returnValue = 0.0;
		if(countoa>0 && countob>0){
			try{
				returnValue = 1-cst.chiSquareTestDataSetsComparison(ArrayUtils.toPrimitive(observeda.toArray(new Long[observeda.size()])), ArrayUtils.toPrimitive(observedb.toArray(new Long[observedb.size()])));
			}catch(NotStrictlyPositiveException e){
				returnValue = 0.0;
			}
		}
		if(returnValue>thresholdP)
			results.add(new DissimilarityResult("UNKOWN", HypothesisType.DIRECT_PRECEDES, returnValue, 0.0));
		
		return results;
	}

	public Set<DissimilarityResult> getEventuallyFollowsDissimilarity() {
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();

		Map<String, Integer> aRow = ef.row(a);
		Map<String, Integer> bRow = ef.row(b);
		List<Long> observeda = new ArrayList<Long>();
		List<Long> observedb = new ArrayList<Long>();
		long countoa = 0;
		long countob = 0;
		for(String c : classifiedNames){
			long oa = aRow.get(c) == null ? 0 : aRow.get(c);
			long ob = bRow.get(c) == null ? 0 : bRow.get(c);
			if(oa>0 || ob>0){
				observeda.add(oa);
				observedb.add(ob);
				countoa += oa;
				countob += ob;
			}
		}
		ChiSquareTest cst = new ChiSquareTest();
		double returnValue = 0.0;
		if(countoa>0 && countob>0){
			try{
				returnValue = 1-cst.chiSquareTestDataSetsComparison(ArrayUtils.toPrimitive(observeda.toArray(new Long[observeda.size()])), ArrayUtils.toPrimitive(observedb.toArray(new Long[observedb.size()])));
			}catch(NotStrictlyPositiveException e){
				returnValue = 0.0;
			}
		}
		
		if(returnValue>thresholdP)
			results.add(new DissimilarityResult("UNKOWN", HypothesisType.EVENTUALLY_FOLLOWS, returnValue, 0.0));
		
		return results;
	}

	public Set<DissimilarityResult> getEventuallyPrecedesDissimilarity() {
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();

		Map<String, Integer> aCol = ef.column(a);
		Map<String, Integer> bCol = ef.column(b);
		List<Long> observeda = new ArrayList<Long>();
		List<Long> observedb = new ArrayList<Long>();
		long countoa = 0;
		long countob = 0;
		for(String c : classifiedNames){
			long oa = aCol.get(c) == null ? 0 : aCol.get(c);
			long ob = bCol.get(c) == null ? 0 : bCol.get(c);
			if(oa>0 || ob>0){
				observeda.add(oa);
				observedb.add(ob);
				countoa += oa;
				countob += ob;
			}
		}
		ChiSquareTest cst = new ChiSquareTest();
		double returnValue = 0.0;
		if(countoa>0 && countob>0){
			try{
				returnValue = 1-cst.chiSquareTestDataSetsComparison(ArrayUtils.toPrimitive(observeda.toArray(new Long[observeda.size()])), ArrayUtils.toPrimitive(observedb.toArray(new Long[observedb.size()])));
			}catch(NotStrictlyPositiveException e){
				returnValue = 0.0;
			}
		}
		
		if(returnValue>thresholdP)
			results.add(new DissimilarityResult("UNKOWN", HypothesisType.DIRECT_PRECEDES, returnValue, 0.0));
		
		return results;
	}
}
