package org.processmining.niektax.temporalrelations.model.dissimilarity;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.google.common.collect.Table;

public abstract class DissimilarityMetric {
	protected String a;
	protected String b;
	protected String originalAb;
	protected String splitOrigin;
	protected Table<String, String, Integer> df;
	protected Table<String, String, Integer> ef;
	protected Table<String, String, Integer> ep;
	protected Table<String, String, Integer> cooc;
	protected Set<String> classifiedNames;
	protected Map<String, Integer> counts;
	protected double thresholdP;
	protected double thresholdOdds;
	
	public DissimilarityMetric(String a, String b, String originalAb, Table<String, String, Integer> df, Table<String, String, Integer> ef, Table<String, String, Integer> ep, Table<String, String, Integer> cooc, Set<String> classifiedNames, Map<String, Integer> counts, double thresholdP, double thresholdOdds){
		this.a = a;
		this.b = b;
		this.originalAb = originalAb;
		this.df = df;
		this.ef = ef;
		this.ep = ep;
		this.cooc = cooc;
		this.classifiedNames = classifiedNames;
		this.counts = counts;
		this.thresholdP = thresholdP;
		this.thresholdOdds = thresholdOdds;
	}
	
	public abstract Set<DissimilarityResult> getParallelDissimilarity();
	public abstract Set<DissimilarityResult> getDirectlyFollowsDissimilarity();
	public abstract Set<DissimilarityResult> getDirectlyPrecedesDissimilarity();
	public abstract Set<DissimilarityResult> getEventuallyFollowsDissimilarity();
	public abstract Set<DissimilarityResult> getEventuallyPrecedesDissimilarity();
	
	public Set<DissimilarityResult> getDissimilarity(){
		Set<DissimilarityResult> s = new HashSet<DissimilarityResult>();
		Set<DissimilarityResult> p = getParallelDissimilarity();
		if(p!=null && !p.isEmpty())
			s.addAll(p);
		Set<DissimilarityResult> df = getDirectlyFollowsDissimilarity();
		if(df!=null && !df.isEmpty())
			s.addAll(df);
		Set<DissimilarityResult> dp = getDirectlyPrecedesDissimilarity();
		if(dp!=null && !dp.isEmpty())
			s.addAll(dp);
		Set<DissimilarityResult> ef = getEventuallyFollowsDissimilarity();
		if(ef!=null && !ef.isEmpty())
			s.addAll(ef);
		Set<DissimilarityResult> ep = getEventuallyPrecedesDissimilarity();
		if(ep!=null && !ep.isEmpty())
			s.addAll(ep);
		return s;
	}
}
