package org.processmining.niektax.temporalrelations.model.dissimilarity;

public class DissimilarityResult implements Comparable<DissimilarityResult>{
	public enum HypothesisType {
		DIRECT_FOLLOWS,
		EVENTUALLY_FOLLOWS,
		DIRECT_PRECEDES,
		EVENTUALLY_PRECEDES,
		PARALLEL
	}
	
	private String otherActivity;
	private HypothesisType type;
	private double p_value;
	private double odds_ratio;

	public DissimilarityResult(String otherActivity, HypothesisType type, double p_value, double odds_ratio){
		this.otherActivity = otherActivity;
		this.type = type;
		this.p_value = p_value;
		this.odds_ratio = odds_ratio;
	}
	
	public String getOtherActivity(){
		return otherActivity;
	}
	
	public HypothesisType getHypothesisType(){
		return type;
	}
	
	public double getPValue(){
		return p_value;
	}
	
	public double getOddsRatio(){
		return odds_ratio;
	}
	
	public boolean testWithBonferroni(long numTests, double threshold){
		return p_value < threshold/numTests;
	}
	
    @Override
    public boolean equals(Object other) {
    	if(other != null && other instanceof DissimilarityResult){
    		DissimilarityResult otherHypothesis = (DissimilarityResult) other;
    		return otherHypothesis.getHypothesisType()==type && otherHypothesis.getOtherActivity().equals(otherActivity);
    	}
        return false;
    }

    @Override
	public int compareTo(DissimilarityResult o) {
		int returnValue = 0;
    	if(o.getOddsRatio()>this.odds_ratio){
			returnValue = -1;
		}else{
			if(o.getOddsRatio()<this.odds_ratio){
				returnValue = 1;
			}
		}
    	return returnValue;
	}
    
    @Override
    public String toString(){
		return "["+otherActivity+", "+type+", "+p_value+", "+odds_ratio+"]";
    }
       
}
