package org.processmining.niektax.temporalrelations.model.dissimilarity;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.commons.math3.stat.inference.AlternativeHypothesis;
import org.apache.commons.math3.stat.inference.BinomialTest;

import com.google.common.collect.Table;

public class ExactBinomialTestDissimilarity extends DissimilarityMetric{

	public ExactBinomialTestDissimilarity(String a, String b, String originalAb, Table<String, String, Integer> df, Table<String, String, Integer> ef, Table<String, String, Integer> ep, Table<String, String, Integer> cooc,Set<String> classifiedNames, Map<String, Integer> counts, double thresholdP, double thresholdOdds){
		super(a,b,originalAb,df,ef,ep,cooc,classifiedNames,counts,thresholdP,thresholdOdds);
	}
	
	public Set<DissimilarityResult> getParallelDissimilarity() {
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();

		// TODO Auto-generated method stub
		return results;
	}

	public Set<DissimilarityResult> getDirectlyFollowsDissimilarity() {
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();

		Integer aCount = counts.get(a) == null ? 0 : counts.get(a);
		Integer bCount = counts.get(b) == null ? 0 : counts.get(b);
		Integer totalEvents = 0;
		for(Integer i : counts.values())
			totalEvents += i;
		
		double maxPValueDiff = 0;
		for(String c : df.columnKeySet()){
			// Follows
			Integer cCount    = counts.get(c) == null ? 0 : counts.get(c);
			Integer acCount   = df.get(a, c) == null ? 0 : df.get(a, c);
			Integer bcCount   = df.get(b, c) == null ? 0 : df.get(b, c);
			double cProbForA = ((double) cCount) / (totalEvents - aCount);
			double cProbForB = ((double) cCount) / (totalEvents - bCount);
			BinomialTest bt = new BinomialTest();
			double aPvalue = bt.binomialTest(aCount, acCount, cProbForA, AlternativeHypothesis.GREATER_THAN);
			double bPvalue = bt.binomialTest(bCount, bcCount, cProbForB, AlternativeHypothesis.GREATER_THAN);
			if(Math.abs(aPvalue-bPvalue)>maxPValueDiff)
				maxPValueDiff = Math.abs(aPvalue-bPvalue);
			

		}
		return null;
	}

	public Set<DissimilarityResult> getDirectlyPrecedesDissimilarity() {
		Integer aCount = counts.get(a) == null ? 0 : counts.get(a);
		Integer bCount = counts.get(b) == null ? 0 : counts.get(b);
		Integer totalEvents = 0;
		for(Integer i : counts.values())
			totalEvents += i;
		
		double maxPValueDiff = 0;
		for(String c : df.columnKeySet()){
			// Preceding
			Integer cCount    = counts.get(c) == null ? 0 : counts.get(c);
			Integer caCount   = df.get(c, a) == null ? 0 : df.get(c, a);
			Integer cbCount   = df.get(c, b) == null ? 0 : df.get(c, b);
			double aProb = ((double) aCount) / (totalEvents - cCount);
			double bProb = ((double) bCount) / (totalEvents - cCount);
			BinomialTest bt = new BinomialTest();
			double aPvalue = bt.binomialTest(cCount, caCount, aProb, AlternativeHypothesis.GREATER_THAN);
			double bPvalue = bt.binomialTest(cCount, cbCount, bProb, AlternativeHypothesis.GREATER_THAN);
			if(Math.abs(aPvalue-bPvalue)>maxPValueDiff)
				maxPValueDiff = Math.abs(aPvalue-bPvalue);
		}
		return null;
	}

	public Set<DissimilarityResult> getEventuallyFollowsDissimilarity() {
		Integer aCount = counts.get(a) == null ? 0 : counts.get(a);
		Integer bCount = counts.get(b) == null ? 0 : counts.get(b);
		Integer totalEvents = 0;
		for(Integer i : counts.values())
			totalEvents += i;
		
		double maxPValueDiff = 0;
		for(String c : ef.columnKeySet()){
			// Follows
			Integer cCount    = counts.get(c) == null ? 0 : counts.get(c);
			Integer acCount   = ef.get(a, c) == null ? 0 : ef.get(a, c);
			Integer bcCount   = ef.get(b, c) == null ? 0 : ef.get(b, c);
			// TODO: this probability is not right... think of way to calculate expected portion of A's eventually followed by C
			// This is hard, because it is dependent on the number of events following A, and therefore trace dependent...
			double cProbForA = ((double) cCount) / (totalEvents - aCount);
			double cProbForB = ((double) cCount) / (totalEvents - bCount);
			BinomialTest bt = new BinomialTest();
			double aPvalue = bt.binomialTest(aCount, acCount, cProbForA, AlternativeHypothesis.GREATER_THAN);
			double bPvalue = bt.binomialTest(bCount, bcCount, cProbForB, AlternativeHypothesis.GREATER_THAN);
			if(Math.abs(aPvalue-bPvalue)>maxPValueDiff)
				maxPValueDiff = Math.abs(aPvalue-bPvalue);
			

		}
		return null;
	}

	public Set<DissimilarityResult> getEventuallyPrecedesDissimilarity() {
		Integer aCount = counts.get(a) == null ? 0 : counts.get(a);
		Integer bCount = counts.get(b) == null ? 0 : counts.get(b);
		Integer totalEvents = 0;
		for(Integer i : counts.values())
			totalEvents += i;
		
		double maxPValueDiff = 0;
		for(String c : ef.columnKeySet()){
			// Preceding
			Integer cCount    = counts.get(c) == null ? 0 : counts.get(c);
			Integer caCount   = ef.get(c, a) == null ? 0 : ef.get(c, a);
			Integer cbCount   = ef.get(c, b) == null ? 0 : ef.get(c, b);
			// TODO: this probability is not right... think of way to calculate expected portion of C's eventually followed by A
			// This is hard, because it is dependent on the number of events following A, and therefore trace dependent...
			double aProb = ((double) aCount) / (totalEvents - cCount);
			double bProb = ((double) bCount) / (totalEvents - cCount);
			BinomialTest bt = new BinomialTest();
			double aPvalue = bt.binomialTest(cCount, caCount, aProb, AlternativeHypothesis.GREATER_THAN);
			double bPvalue = bt.binomialTest(cCount, cbCount, bProb, AlternativeHypothesis.GREATER_THAN);
			if(Math.abs(aPvalue-bPvalue)>maxPValueDiff)
				maxPValueDiff = Math.abs(aPvalue-bPvalue);
		}
		return null;
	}

}
