package org.processmining.niektax.temporalrelations.model.dissimilarity;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.processmining.niektax.temporalrelations.model.dissimilarity.DissimilarityResult.HypothesisType;

import com.google.common.collect.Table;

import edu.northwestern.at.utils.math.statistics.FishersExactTest;

public class FishersExactTestDissimilarity extends DissimilarityMetric{

	public FishersExactTestDissimilarity(String a, String b, String originalAb, Table<String, String, Integer> df, Table<String, String, Integer> ef, Table<String, String, Integer> ep, Table<String, String, Integer> cooc, Set<String> classifiedNames, Map<String, Integer> counts, double thresholdP, double thresholdOdds){
		super(a,b,originalAb,df,ef,ep,cooc,classifiedNames,counts,thresholdP,thresholdOdds);
	}
	
	public Set<DissimilarityResult> getParallelDissimilarity() {
		
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();
		/*
		Integer aCount = counts.get(a) == null ? 0 : counts.get(a);
		Integer bCount = counts.get(b) == null ? 0 : counts.get(b);
		for(String c : ef.columnKeySet()){
			Integer a_cooc_c  = cooc.get(a, c) == null ? 0 : cooc.get(a, c);
			Integer b_cooc_c  = cooc.get(b, c) == null ? 0 : cooc.get(b, c);
			
			if(c.equals(a)||c.equals(b)||c.equals(originalAb))
				continue;
			
			Integer a_ncooc_c = aCount-a_cooc_c;
			Integer b_ncooc_c = bCount-b_cooc_c;
			
			//System.out.println("a: "+a+" b: "+b+" c: "+c+" a_cooc_c: "+a_cooc_c+" b_cooc_c: "+b_cooc_c+" a_ncooc_c: "+a_ncooc_c+" b_ncooc_c: "+b_ncooc_c);
			
			double p_dissimilarity = 1-FishersExactTest.fishersExactTest(a_cooc_c, b_cooc_c, a_ncooc_c, b_ncooc_c)[0];

			if(p_dissimilarity>thresholdP){
				double odds_ratio = oddsRatio(a_cooc_c, b_cooc_c, a_ncooc_c, b_ncooc_c);
				if(odds_ratio>thresholdOdds)
					results.add(new DissimilarityResult(c, HypothesisType.PARALLEL, p_dissimilarity, odds_ratio));
			}
		}
		*/
		return results;
	}

	private static double oddsRatio(int a, int b, int c, int d){
		double first  = a * d;
		double second = b * c;
		double oddsRatio;
		if(first>second){
			if(second>0)
				oddsRatio = first/second;
			else
				oddsRatio = Double.NaN;
		}else{
			if(first>0)
				oddsRatio = second/first;
			else
				oddsRatio = Double.NaN;
		}
		return oddsRatio;
	}
	
	public Set<DissimilarityResult> getDirectlyFollowsDissimilarity() {
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();
		
		Integer aCount = counts.get(a) == null ? 0 : counts.get(a);
		Integer bCount = counts.get(b) == null ? 0 : counts.get(b);
		for(String c : df.columnKeySet()){
			if(c.equals(a)||c.equals(b)||c.equals(originalAb)||c.equals(splitOrigin))
				continue;
			
			Integer a_df_c  = df.get(a, c) == null ? 0 : df.get(a, c);
			Integer b_df_c = df.get(b, c) == null ? 0 : df.get(b, c);

			Integer a_ndf_c = aCount-a_df_c;
			Integer b_ndf_c = bCount-b_df_c;
			
			double p_dissimilarity = FishersExactTest.fishersExactTest(a_df_c, b_df_c, a_ndf_c, b_ndf_c)[0];
			if(p_dissimilarity<thresholdP){
				double odds_ratio = oddsRatio(a_df_c, b_df_c, a_ndf_c, b_ndf_c);
				if(odds_ratio>thresholdOdds)
					results.add(new DissimilarityResult(c, HypothesisType.DIRECT_FOLLOWS, p_dissimilarity, odds_ratio));
			}

		}
		
		return results;	
	}

	public Set<DissimilarityResult> getDirectlyPrecedesDissimilarity() {
		
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();
		
		Integer aCount = counts.get(a) == null ? 0 : counts.get(a);
		Integer bCount = counts.get(b) == null ? 0 : counts.get(b);
		
		for(String c : df.columnKeySet()){
			Integer a_dp_c    = df.get(c, a) == null ? 0 : df.get(c, a);
			Integer b_dp_c    = df.get(c, b) == null ? 0 : df.get(c, b);
			
			if(c.equals(a)||c.equals(b)){
				a_dp_c = (df.get(a, a) == null ? 0 : df.get(a, a)) + (df.get(b, a) == null ? 0 : df.get(b, a));
				b_dp_c = (df.get(a, b) == null ? 0 : df.get(a, b)) + (df.get(b, b) == null ? 0 : df.get(b, b));
			}
			
			Integer a_ndp_c = aCount-a_dp_c;
			Integer b_ndp_c = bCount-b_dp_c;
			
			double p_dissimilarity = FishersExactTest.fishersExactTest(a_dp_c, b_dp_c, a_ndp_c, b_ndp_c)[0];
			if(p_dissimilarity<thresholdP){
				double odds_ratio = oddsRatio(a_dp_c, b_dp_c, a_ndp_c, b_ndp_c);
				if(odds_ratio>thresholdOdds)
					results.add(new DissimilarityResult(c, HypothesisType.DIRECT_PRECEDES, p_dissimilarity, odds_ratio));
			}
		}
		
		return results;
	}

	public Set<DissimilarityResult> getEventuallyFollowsDissimilarity() {

		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();
		
		Integer aCount = counts.get(a) == null ? 0 : counts.get(a);
		Integer bCount = counts.get(b) == null ? 0 : counts.get(b);
		for(String c : ef.columnKeySet()){
			Integer a_ef_c  = ef.get(a, c) == null ? 0 : ef.get(a, c);
			Integer b_ef_c  = ef.get(b, c) == null ? 0 : ef.get(b, c);

			if(c.equals(a)||c.equals(b)||c.equals(originalAb))
				continue; // EVENTUALLY is not fixable by adding c=a and c=b, as we would then double-count
			
			Integer a_nef_c = aCount-a_ef_c;
			Integer b_nef_c = bCount-b_ef_c;
			
			double p_dissimilarity = FishersExactTest.fishersExactTest(a_ef_c, b_ef_c, a_nef_c, b_nef_c)[0];

			if(p_dissimilarity<thresholdP){
				double odds_ratio = oddsRatio(a_ef_c, b_ef_c, a_nef_c, b_nef_c);
				if(odds_ratio>thresholdOdds)
					results.add(new DissimilarityResult(c, HypothesisType.EVENTUALLY_FOLLOWS, p_dissimilarity, odds_ratio));
			}
		}
		
		return results;
	}

	public Set<DissimilarityResult> getEventuallyPrecedesDissimilarity() {
		Set<DissimilarityResult> results = new HashSet<DissimilarityResult>();

		Integer aCount = counts.get(a) == null ? 0 : counts.get(a);
		Integer bCount = counts.get(b) == null ? 0 : counts.get(b);
		
		for(String c : ep.columnKeySet()){
			Integer a_ep_c  = ep.get(a, c) == null ? 0 : ep.get(a, c);
			Integer b_ep_c  = ep.get(b, c) == null ? 0 : ep.get(b, c);
			
			if(c.equals(a)||c.equals(b))
				continue; // EVENTUALLY is not fixable by adding c=a and c=b, as we would then double-count

			Integer a_nep_c = aCount-a_ep_c;
			Integer b_nep_c = bCount-b_ep_c;
			double p_dissimilarity = FishersExactTest.fishersExactTest(a_ep_c, b_ep_c, a_nep_c, b_nep_c)[0];
			if(p_dissimilarity<thresholdP){
				double odds_ratio = oddsRatio(a_ep_c, b_ep_c, a_nep_c, b_nep_c);
				if(odds_ratio>thresholdOdds)
					results.add(new DissimilarityResult(c, HypothesisType.EVENTUALLY_PRECEDES, p_dissimilarity, odds_ratio));
			}
		}
		return results;
	}

}
